import asyncio
import logging
import re
from typing import Optional

from irc.modes import parse_channel_modes

from heisenbridge.channel_room import ChannelRoom
from heisenbridge.command_parse import CommandParser
from heisenbridge.matrix import MatrixError
from heisenbridge.private_room import parse_irc_formatting
from heisenbridge.room import unpack_member_states


class NetworkRoom:
    pass


def connected(f):
    def wrapper(*args, **kwargs):
        self = args[0]

        if not self.network or not self.network.conn or not self.network.conn.connected:
            return asyncio.sleep(0)

        return f(*args, **kwargs)

    return wrapper


class PlumbedRoom(ChannelRoom):
    need_invite = False
    max_lines = 5
    use_pastebin = True
    use_displaynames = True
    use_disambiguation = True
    use_zwsp = False
    allow_notice = False
    force_forward = True
    topic_sync = None

    def init(self) -> None:
        super().init()

        cmd = CommandParser(
            prog="MAXLINES", description="set maximum number of lines per message until truncation or pastebin"
        )
        cmd.add_argument("lines", type=int, nargs="?", help="Number of lines")
        self.commands.register(cmd, self.cmd_maxlines)

        cmd = CommandParser(prog="PASTEBIN", description="enable or disable automatic pastebin of long messages")
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable pastebin")
        cmd.add_argument(
            "--disable", dest="enabled", action="store_false", help="Disable pastebin (messages will be truncated)"
        )
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_pastebin)

        cmd = CommandParser(
            prog="DISPLAYNAMES", description="enable or disable use of displaynames in relayed messages"
        )
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable displaynames")
        cmd.add_argument(
            "--disable", dest="enabled", action="store_false", help="Disable displaynames (fallback to MXID)"
        )
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_displaynames)

        cmd = CommandParser(
            prog="DISAMBIGUATION", description="enable or disable disambiguation of conflicting displaynames"
        )
        cmd.add_argument(
            "--enable", dest="enabled", action="store_true", help="Enable disambiguation (postfix with MXID)"
        )
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable disambiguation")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_disambiguation)

        cmd = CommandParser(prog="ZWSP", description="enable or disable Zero-Width-Space anti-ping")
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable ZWSP anti-ping")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable ZWSP anti-ping")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_zwsp)

        cmd = CommandParser(prog="NOTICERELAY", description="enable or disable relaying of Matrix notices to IRC")
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable notice relay")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable notice relay")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_noticerelay)

        cmd = CommandParser(prog="TOPIC", description="show or set channel topic and configure sync mode")
        cmd.add_argument("--sync", choices=["off", "irc", "matrix", "any"], help="Topic sync targets, defaults to off")
        cmd.add_argument("text", nargs="*", help="topic text if setting")
        self.commands.register(cmd, self.cmd_topic)

        self.mx_register("m.room.topic", self._on_mx_room_topic)

    def is_valid(self) -> bool:
        # we are valid as long as the appservice is in the room
        if not self.in_room(self.serv.user_id):
            return False

        return True

    @staticmethod
    async def create(network: "NetworkRoom", id: str, channel: str, key: str) -> "ChannelRoom":
        logging.debug(f"PlumbedRoom.create(network='{network.name}', id='{id}', channel='{channel}', key='{key}'")

        try:
            resp = await network.serv.api.post_room_join_alias(id)
            join_rules = await network.serv.api.get_room_state_event(resp["room_id"], "m.room.join_rules")
            members = await network.serv.api.get_room_members(resp["room_id"])
        except MatrixError as e:
            network.send_notice(f"Failed to join room: {str(e)}")
            return

        joined, banned = unpack_member_states(members)

        room = PlumbedRoom(resp["room_id"], network.user_id, network.serv, joined.keys(), banned.keys())
        room.name = channel.lower()
        room.key = key
        room.network = network
        room.network_id = network.id
        room.network_name = network.name
        room.need_invite = join_rules["join_rule"] != "public"

        # stamp global member sync setting at room creation time
        room.member_sync = network.serv.config["member_sync"]

        for user_id, displayname in joined.items():
            if displayname is not None:
                room.displaynames[user_id] = displayname

        network.serv.register_room(room)
        network.rooms[room.name] = room
        await room.save()

        network.send_notice(f"Plumbed {resp['room_id']} to {channel}, to unplumb just kick me out.")
        return room

    def from_config(self, config: dict) -> None:
        super().from_config(config)

        if "max_lines" in config:
            self.max_lines = config["max_lines"]

        if "use_pastebin" in config:
            self.use_pastebin = config["use_pastebin"]

        if "use_displaynames" in config:
            self.use_displaynames = config["use_displaynames"]

        if "use_disambiguation" in config:
            self.use_disambiguation = config["use_disambiguation"]

        if "use_zwsp" in config:
            self.use_zwsp = config["use_zwsp"]

        if "allow_notice" in config:
            self.allow_notice = config["allow_notice"]

        if "topic_sync" in config:
            self.topic_sync = config["topic_sync"]

    def to_config(self) -> dict:
        return {
            **(super().to_config()),
            "max_lines": self.max_lines,
            "use_pastebin": self.use_pastebin,
            "use_displaynames": self.use_displaynames,
            "use_disambiguation": self.use_disambiguation,
            "use_zwsp": self.use_zwsp,
            "allow_notice": self.allow_notice,
            "topic_sync": self.topic_sync,
        }

    # topic updates from channel state replies are ignored because formatting changes
    def set_topic(self, topic: str, user_id: Optional[str] = None) -> None:
        pass

    def on_topic(self, conn, event) -> None:
        self.send_notice("{} changed the topic".format(event.source.nick))
        if conn.real_nickname != event.source.nick and self.topic_sync in ["matrix", "any"]:
            (plain, formatted) = parse_irc_formatting(event.arguments[0])
            super().set_topic(plain)

    @connected
    async def _on_mx_room_topic(self, event) -> None:
        if event["sender"] != self.serv.user_id and self.topic_sync in ["irc", "any"]:
            topic = re.sub(r"[\r\n]", " ", event["content"]["topic"])
            self.network.conn.topic(self.name, topic)

    @connected
    async def on_mx_message(self, event) -> None:
        sender = event["sender"]
        (name, server) = sender.split(":")

        # ignore self messages
        if sender == self.serv.user_id:
            return

        # prevent re-sending federated messages back
        if name.startswith("@" + self.serv.puppet_prefix) and server == self.serv.server_name:
            return

        # add ZWSP to sender to avoid pinging on IRC
        if self.use_zwsp:
            sender = f"{name[:2]}\u200B{name[2:]}:{server[:1]}\u200B{server[1:]}"

        if self.use_displaynames and event["sender"] in self.displaynames:
            sender_displayname = self.displaynames[event["sender"]]

            # ensure displayname is unique
            if self.use_disambiguation:
                for user_id, displayname in self.displaynames.items():
                    if user_id != event["sender"] and displayname == sender_displayname:
                        sender_displayname += f" ({sender})"
                        break

            # add ZWSP if displayname matches something on IRC
            if self.use_zwsp and len(sender_displayname) > 1:
                sender_displayname = f"{sender_displayname[:1]}\u200B{sender_displayname[1:]}"

            sender = sender_displayname

        # limit plumbed sender max length to 100 characters
        sender = sender[:100]

        if event["content"]["msgtype"] in ["m.image", "m.file", "m.audio", "m.video"]:

            # process media event like it was a text message
            media_event = {"content": {"body": self.serv.mxc_to_url(event["content"]["url"], event["content"]["body"])}}
            messages = self._process_event_content(media_event, prefix=f"<{sender}> ")
            self.network.conn.privmsg(self.name, messages[0])

            self.react(event["event_id"], "\U0001F517")  # link
            self.media.append([event["event_id"], event["content"]["url"]])
            await self.save()
        elif event["content"]["msgtype"] == "m.emote":
            await self._send_message(event, self.network.conn.action, prefix=f"{sender} ")
        elif event["content"]["msgtype"] == "m.text":
            await self._send_message(event, self.network.conn.privmsg, prefix=f"<{sender}> ")
        elif event["content"]["msgtype"] == "m.notice" and self.allow_notice:
            await self._send_message(event, self.network.conn.notice, prefix=f"<{sender}> ")

        await self.serv.api.post_room_receipt(event["room_id"], event["event_id"])

    @connected
    async def on_mx_ban(self, user_id) -> None:
        nick = self.serv.nick_from_irc_user_id(self.network.name, user_id)
        if nick is None:
            return

        # best effort kick and ban
        self.network.conn.mode(self.name, f"+b {nick}!*@*")
        self.network.conn.kick(self.name, nick, "You have been banned on Matrix")

    @connected
    async def on_mx_unban(self, user_id) -> None:
        nick = self.serv.nick_from_irc_user_id(self.network.name, user_id)
        if nick is None:
            return

        # best effort unban
        self.network.conn.mode(self.name, f"-b {nick}!*@*")

    @connected
    async def on_mx_leave(self, user_id) -> None:
        nick = self.serv.nick_from_irc_user_id(self.network.name, user_id)
        if nick is None:
            return

        # best effort kick
        if self.is_on_channel(nick):
            self.network.conn.kick(self.name, nick, "You have been kicked on Matrix")

    def pills(self):
        ret = super().pills()

        # remove the bot from pills as it may cause confusion
        nick = self.network.conn.real_nickname.lower()
        if nick in ret:
            del ret[nick]

        return ret

    async def cmd_maxlines(self, args) -> None:
        if args.lines is not None:
            self.max_lines = args.lines
            await self.save()

        self.send_notice(f"Max lines is {self.max_lines}")

    async def cmd_pastebin(self, args) -> None:
        if args.enabled is not None:
            self.use_pastebin = args.enabled
            await self.save()

        self.send_notice(f"Pastebin is {'enabled' if self.use_pastebin else 'disabled'}")

    async def cmd_displaynames(self, args) -> None:
        if args.enabled is not None:
            self.use_displaynames = args.enabled
            await self.save()

        self.send_notice(f"Displaynames are {'enabled' if self.use_displaynames else 'disabled'}")

    async def cmd_disambiguation(self, args) -> None:
        if args.enabled is not None:
            self.use_disambiguation = args.enabled
            await self.save()

        self.send_notice(f"Dismabiguation is {'enabled' if self.use_disambiguation else 'disabled'}")

    async def cmd_zwsp(self, args) -> None:
        if args.enabled is not None:
            self.use_zwsp = args.enabled
            await self.save()

        self.send_notice(f"Zero-Width-Space anti-ping is {'enabled' if self.use_zwsp else 'disabled'}")

    async def cmd_noticerelay(self, args) -> None:
        if args.enabled is not None:
            self.allow_notice = args.enabled
            await self.save()

        self.send_notice(f"Notice relay is {'enabled' if self.allow_notice else 'disabled'}")

    async def cmd_topic(self, args) -> None:
        if args.sync is None:
            self.network.conn.topic(self.name, " ".join(args.text))
            return

        self.topic_sync = args.sync if args.sync != "off" else None
        self.send_notice(f"Topic sync is {self.topic_sync if self.topic_sync else 'off'}")
        await self.save()

    def on_mode(self, conn, event) -> None:
        super().on_mode(conn, event)

        # when we get ops (or half-ops) get current ban list to see if we need to ban someone that has been banned on matrix
        modes = list(event.arguments)
        for sign, key, value in parse_channel_modes(" ".join(modes)):
            if sign == "+" and key in ["o", "h"] and value == self.network.conn.real_nickname:
                self.network.conn.mode(self.name, "+b")

    def on_endofbanlist(self, conn, event) -> None:
        masks = [ban[0].lower() for ban in self.bans_buffer]
        super().on_endofbanlist(conn, event)

        # add any nick bans that are missing from IRC
        for user_id in self.bans:
            nick = self.serv.nick_from_irc_user_id(self.network.name, user_id)
            if nick is None:
                continue

            mask = f"{nick}!*@*"
            if mask not in masks:
                self.network.conn.mode(self.name, f"+b {mask}")
            if self.is_on_channel(nick):
                self.network.conn.kick(self.name, nick)



import asyncio
import logging
import time
import urllib

from aiohttp import ClientError
from aiohttp import ClientResponseError
from aiohttp import ClientSession
from aiohttp import TCPConnector


class MatrixError(Exception):
    def __init__(self, data):
        if "errcode" in data:
            self.errcode = data["errcode"]
        else:
            self.errcode = 0

        if "error" in data:
            self.error = data["error"]
        else:
            self.error = "Unspecified error"

        super().__init__(self.errcode)


class MatrixErrorUnknown(MatrixError):
    pass


class MatrixNotFound(MatrixError):
    pass


class MatrixForbidden(MatrixError):
    pass


class MatrixUserInUse(MatrixError):
    pass


class MatrixLimitExceeded(MatrixError):
    def __init__(self, data):
        super().__init__(data)

        if "retry_after_ms" in data:
            self.retry_after_s = data["retry_after_ms"] / 1000
        else:
            self.retry_after_s = 5


class Matrix:
    def __init__(self, url, token):
        self.url = url
        self.token = token
        self.seq = 0
        self.session = str(int(time.time()))
        self.conn = TCPConnector()

    def _matrix_error(self, data):
        errors = {
            "M_UNKNOWN": MatrixErrorUnknown,
            "M_NOT_FOUND": MatrixNotFound,
            "M_FORBIDDEN": MatrixForbidden,
            "M_USER_IN_USE": MatrixUserInUse,
            "M_LIMIT_EXCEEDED": MatrixLimitExceeded,
        }

        ex = errors.get(data["errcode"], MatrixError)
        return ex(data)

    def _txn(self):
        self.seq += 1
        return self.session + "-" + str(self.seq)

    async def call(self, method, uri, data=None, content_type="application/json", retry=True):
        if method == "POST" and data is None:
            data = {}

        async with ClientSession(
            headers={"Authorization": "Bearer " + self.token}, connector=self.conn, connector_owner=False
        ) as session:
            for i in range(0, 60):
                try:
                    if content_type == "application/json":
                        resp = await session.request(method, self.url + uri, json=data)
                    else:
                        resp = await session.request(
                            method, self.url + uri, data=data, headers={"Content-type": content_type}
                        )
                    ret = await resp.json()

                    if resp.status > 299:
                        raise self._matrix_error(ret)

                    return ret
                except MatrixErrorUnknown:
                    logging.warning(
                        f"Request to HS failed with unknown Matrix error, HTTP code {resp.status}, falling through to retry."
                    )
                except MatrixLimitExceeded as e:
                    logging.warning(f"Request to HS was rate limited, retrying in {e.retry_after_s} seconds...")
                    await asyncio.sleep(e.retry_after_s)
                    continue
                except ClientResponseError as e:
                    # fail fast if no retry allowed if dealing with HTTP error
                    logging.debug(str(e))
                    if not retry:
                        raise

                except (ClientError, asyncio.TimeoutError) as e:
                    # catch and fall-through to sleep
                    logging.debug(str(e))
                    pass

                logging.warning(f"Request to HS failed, assuming it is down, retry {i+1}/60...")
                await asyncio.sleep(30)

    async def get_user_whoami(self):
        return await self.call("GET", "/_matrix/client/r0/account/whoami")

    async def get_user_joined_rooms(self):
        return await self.call("GET", "/_matrix/client/r0/joined_rooms")

    async def get_user_account_data(self, user_id, key):
        user_id = urllib.parse.quote(user_id, safe="")
        return await self.call("GET", "/_matrix/client/r0/user/" + user_id + "/account_data/" + key)

    async def put_user_account_data(self, user_id, key, data):
        user_id = urllib.parse.quote(user_id, safe="")
        return await self.call("PUT", "/_matrix/client/r0/user/" + user_id + "/account_data/" + key, data)

    async def get_room_account_data(self, user_id, room_id, key):
        user_id = urllib.parse.quote(user_id, safe="")
        return await self.call(
            "GET",
            "/_matrix/client/r0/user/" + user_id + "/rooms/" + room_id + "/account_data/" + key,
        )

    async def put_room_account_data(self, user_id, room_id, key, data):
        user_id = urllib.parse.quote(user_id, safe="")
        return await self.call(
            "PUT",
            "/_matrix/client/r0/user/" + user_id + "/rooms/" + room_id + "/account_data/" + key,
            data,
        )

    async def post_room_leave(self, room_id, user_id=None):
        if user_id:
            user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "POST",
            "/_matrix/client/r0/rooms/" + room_id + "/leave" + ("?user_id={}".format(user_id) if user_id else ""),
        )

    async def post_room_kick(self, room_id, target_user_id, reason="", user_id=None):
        if user_id:
            user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "POST",
            "/_matrix/client/r0/rooms/" + room_id + "/kick" + ("?user_id={}".format(user_id) if user_id else ""),
            {
                "reason": reason,
                "user_id": target_user_id,
            },
        )

    async def post_room_forget(self, room_id):
        return await self.call("POST", "/_matrix/client/r0/rooms/" + room_id + "/forget")

    async def get_room_joined_members(self, room_id):
        return await self.call("GET", "/_matrix/client/r0/rooms/" + room_id + "/joined_members")

    async def get_room_members(self, room_id, not_membership="leave"):
        q = ""
        if not_membership:
            q = f"?not_membership={not_membership}"
        return await self.call("GET", "/_matrix/client/r0/rooms/" + room_id + "/members" + q)

    async def get_room_event(self, room_id, event_id):
        return await self.call("GET", "/_matrix/client/r0/rooms/" + room_id + "/event/" + event_id)

    async def get_room_state_event(self, room_id, event_type, state_key=""):
        return await self.call("GET", "/_matrix/client/r0/rooms/" + room_id + "/state/" + event_type + "/" + state_key)

    async def post_room_join(self, room_id, user_id=None):
        if user_id:
            user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "POST",
            "/_matrix/client/r0/rooms/" + room_id + "/join" + ("?user_id={}".format(user_id) if user_id else ""),
        )

    async def post_room_join_alias(self, room_alias, user_id=None):
        server_name = room_alias.split(":")[1]
        room_alias = urllib.parse.quote(room_alias, safe="")
        if user_id:
            user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "POST",
            f"/_matrix/client/r0/join/{room_alias}?server_name={server_name}"
            + ("&user_id={}".format(user_id) if user_id else ""),
        )

    async def post_room_invite(self, room_id, user_id):
        return await self.call(
            "POST",
            "/_matrix/client/r0/rooms/" + room_id + "/invite",
            {"user_id": user_id},
        )

    async def put_room_send_event(self, room_id, type, content, user_id=None):
        if user_id:
            user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "PUT",
            "/_matrix/client/r0/rooms/"
            + room_id
            + "/send/"
            + type
            + "/"
            + self._txn()
            + ("?user_id={}".format(user_id) if user_id else ""),
            content,
        )

    async def put_room_send_state(self, room_id, type, state_key, content, user_id=None):
        if user_id:
            user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "PUT",
            "/_matrix/client/r0/rooms/"
            + room_id
            + "/state/"
            + type
            + "/"
            + state_key
            + ("?user_id={}".format(user_id) if user_id else ""),
            content,
        )

    async def post_room_create(self, data):
        return await self.call("POST", "/_matrix/client/r0/createRoom", data)

    async def post_room_receipt(self, room_id, event_id, receipt_type="m.read"):
        room_id = urllib.parse.quote(room_id, safe="")
        event_id = urllib.parse.quote(event_id, safe="")
        receipt_type = urllib.parse.quote(receipt_type, safe="")

        return await self.call("POST", f"/_matrix/client/r0/rooms/{room_id}/receipt/{receipt_type}/{event_id}")

    async def post_user_register(self, data):
        return await self.call("POST", "/_matrix/client/r0/register?kind=user", data)

    async def put_user_displayname(self, user_id, displayname):
        user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "PUT",
            "/_matrix/client/r0/profile/{}/displayname?user_id={}".format(user_id, user_id),
            {"displayname": displayname},
        )

    async def put_user_avatar_url(self, user_id, url):
        user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "PUT",
            "/_matrix/client/r0/profile/{}/avatar_url?user_id={}".format(user_id, user_id),
            {"avatar_url": url},
        )

    async def get_user_avatar_url(self, user_id):
        user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "GET",
            "/_matrix/client/r0/profile/{}/avatar_url?user_id={}".format(user_id, user_id),
        )

    async def put_user_presence(self, user_id, presence="online", status_msg=""):
        user_id = urllib.parse.quote(user_id, safe="")

        return await self.call(
            "PUT", f"/_matrix/client/r0/presence/{user_id}/status", {"presence": presence, "status_msg": status_msg}
        )

    async def post_media_upload(self, data, content_type, filename=None):
        return await self.call(
            "POST",
            "/_matrix/media/r0/upload" + ("?filename=" + urllib.parse.quote(filename, safe="") if filename else ""),
            data,
            content_type=content_type,
        )

    async def get_synapse_admin_users_admin(self, user_id):
        user_id = urllib.parse.quote(user_id, safe="")
        return await self.call("GET", f"/_synapse/admin/v1/users/{user_id}/admin", retry=False)

    async def post_synapse_admin_room_join(self, room_id, user_id):
        return await self.call("POST", f"/_synapse/admin/v1/join/{room_id}", {"user_id": user_id})

    async def post_synapse_admin_media_quarantine(self, server_name, media_id):
        server_name = urllib.parse.quote(server_name, safe="")
        media_id = urllib.parse.quote(media_id, safe="")
        return await self.call("POST", f"/_synapse/admin/v1/media/quarantine/{server_name}/{media_id}")
